package dziuba.reflect;

import java.lang.reflect.*;
import java.util.*;

/**
 * Propery mapper.
 * Retrieves property getters and setters
 * from bean instance and get access to them.
 * Getter must be:
 *      type getPropertyName() or
 *   boolean  isPropertyName()
 * Setter must be:
 *      void setPropertyName(type)
 *
 * TODO: property name transformations (firstSecond, first_second, first-second)
 *
 * @author Alexey Dsiuba
 */
public class PropertyMapper {

    public static final int NONE   = 0;
    public static final int ANY    = 0;
    public static final int GETTER = 1;
    public static final int SETTER = 2;
    public static final int BOTH   = GETTER | SETTER;

    /**
     * Creates PropertyMapper for bean instance.
     * @throws NullpointerException if bean == null
     */
    public PropertyMapper(Object bean) {
        this.bean = bean;
        Method[] methods = bean.getClass().getMethods();
        for (int i = 0; i < methods.length; i++) {
            if (isGetter(methods[i])) {
                String name = methods[i].getName().substring(
                    methods[i].getName().startsWith("get") ? 3 : 2);
                if (name.length() > 0)
                    getters.put(transformName(name), methods[i]);
            }
            else if (isSetter(methods[i])) {
                String name = methods[i].getName().substring(3);
                if (name.length() > 0)
                    setters.put(transformName(name), methods[i]);
            }
        }
        rwNames.addAll(setters.keySet());
        rwNames.retainAll(getters.keySet());
    }
    
    /**
     * Checks wether bean has such property
     * and returns property access.
     * assert(pm.hasProperty("propertyName") & GETTER) -- has getter for it
     * assert(pm.hasProperty("propertyName") != NONE)  -- has property
     *
     * @returns NONE, GETTER, SETTER or BOTH
     */
    public int hasProperty(String property) {
        return (setters.containsKey(property) ? SETTER : NONE)
             | (getters.containsKey(property) ? GETTER : NONE);
    }

    /**
     * Gets property type or null if no such property found.
     */
    public Class getPropertyType(String name) {
        return setters.containsKey(name)
            ? setters.get(name).getParameterTypes()[0]
            : (getters.containsKey(name)
                ? getters.get(name).getReturnType() : null);
    }

    /**
     * Gets bean property names.
     * @param access requested property filter
     *               can be GETTER, SETTER, BOTH or ANY
     */    
    public Set<String> getPropertyNames(int access) {
        Set<String> properties = new HashSet<String>();
        if (access == SETTER || access == ANY)
            properties.addAll(setters.keySet());
        if (access == GETTER || access == ANY)
            properties.addAll(getters.keySet());
        if (access == BOTH)
            properties.addAll(rwNames);
        return properties;
    }
    
    /**
     * Sets bean property value.
     * 
     * @throws NullpointerException if no setter for property exists
     * @throws IllegalArgumentException if wrong value type
     * @throws Throwable if setter throws an exception
     */
    public void set(String property, Object value) throws Throwable {
        try {
            setters.get(property).invoke(bean, value);
        }
        catch (IllegalAccessException e) {
            throw new RuntimeException(e);
        }
        catch (InvocationTargetException e) {
            throw e.getCause();
        }
    }

    /**
     * Gets bean property value.
     * 
     * @throws NullpointerException if no setter for property exists
     * @throws Throwable if getter throws an exception
     */
    public Object get(String property) throws Throwable {
        try {
            return getters.get(property).invoke(bean);
        }
        catch (IllegalAccessException e) {
            throw new RuntimeException(e);
        }
        catch (InvocationTargetException e) {
            throw e.getCause();
        }
    }

    public String toString() {
        StringBuffer s = new StringBuffer(getClass().getName());
        return s.append("@").append(bean).toString();
    }
    
    /**
     * Returns {name:value} map of bean properties.
     * @throws Exception if getter throws an exception,
     *                   message contains property, cause - exception
     */
    public Map<String,Object> getProperties() throws Exception {
        String name = null;
        try {
            Map<String,Object> values = new HashMap<String,Object>();
            Iterator<String> names = getters.keySet().iterator();
            while (names.hasNext()) {
                name = names.next();
                values.put(name, get(name));
            }
            return values;
        }
        catch (Throwable e) {
            throw new Exception(name, e);
        }
    }

    /**
     * Updates properties from map.
     * @throws Exception if setter throws an exception,
     *                   or if type mismatch 
     *                   (Exception.getCause() == IllegalArgumentException)
     */
    public void setProperties(Map<String,Object> values) throws Exception {
        String name = null;
        try {
            Iterator<String> names = values.keySet().iterator();
            while (names.hasNext()) {
                name = names.next();
                if (setters.containsKey(name))
                    set(name, values.get(name));
            }
        }
        catch (Throwable e) {
            throw new Exception(name, e);
        }
    }
    
    // static:
    
    public static boolean isGetter(Method method) {
        return (method.getName().startsWith("get")
            || (method.getName().startsWith("is")
                && method.getReturnType() == Boolean.TYPE))
            && method.getParameterTypes().length == 0;
    }
    
    public static boolean isSetter(Method method) {
        return method.getName().startsWith("set")
            && method.getReturnType() == Void.TYPE
            && method.getParameterTypes().length == 1;
    }

    private static String transformName(String name) {
        return name.substring(0,1).toLowerCase() + name.substring(1);
    }

    // private:
    
    private Map<String,Method> getters = new HashMap<String,Method>();
    private Map<String,Method> setters = new HashMap<String,Method>();

    private Set<String> rwNames = new HashSet<String>();
    
    private Object bean;
}
